VERSION 1.0 CLASS
BEGIN
  MultiUse = -1  'True
  Persistable = 0  'NotPersistable
  DataBindingBehavior = 0  'vbNone
  DataSourceBehavior  = 0  'vbNone
  MTSTransactionMode  = 0  'NotAnMTSObject
END
Attribute VB_Name = "InternalDictionary"
Attribute VB_GlobalNameSpace = False
Attribute VB_Creatable = False
Attribute VB_PredeclaredId = False
Attribute VB_Exposed = False
Option Explicit

Public Enum CollectionAddOption
    AddBeforePosition = 1&
    AddAfterPosition = 2&
    BeginningOfCollection = 3&
    EndOfCollection = 4&
End Enum

Public Enum CollectionError
    AlreadyInitialized = &H80000500
    KeyNotFound = &H80000501
    MoveNotAllowedWhileInEnum = &H80000502
    InvalidDataType = &H80000503
    NoPositionSupplied = &H80000504
    Undefined = &H80000505
End Enum

Private Type CollectionKey
    bInitialized As Boolean
    sKey As String
    nIndex As Long
End Type

'local array for fast key lookups
Private m_aKeys() As CollectionKey

' cross lookup array used to lookup items from keys as well as
' being copied into enumerator objects
Private m_nHoldKeyArrayIndex As Long

' holds the current number of enumeration objects in effect for this instance of the collection
Private m_nCurrentEnumOperations As Long

Public Enum CompareMode
    BinaryCompare = vbBinaryCompare
    TextCompare = vbTextCompare
End Enum

' change this value if you want your keys to be case sensitive
Private m_eCompareMode As CompareMode

' allows the property set for the item property to be activated
Private m_bAllowItemAsignments As Boolean
Private m_bDisallowKeyCreation As Boolean

' change this if you want something other than a one based collection
Private m_nBaseIndex As Long

' since the collection can start at any number (including negitive numbers)
' we need a value that can be returned when a key was not found
Private Const KEY_NOT_FOUND As Long = (-&HEFFFFFFF)

' array used to lookup indexes in the m_aItems array
Private m_anIndexLookp() As Long

'Public Type CollectionValueStruct
'    bInitialized As Boolean
'    ' set this member to the type of object you are collecting
'    oItem As Variant
'    sKey As String
'    bDirty As Boolean
'End Type

'local variable to hold collection
Private m_aItems() As CollectionValueStruct

'Public Event GenericEvent(ByVal Index As Long, ByVal Key As String)

Private Declare Sub RtlMoveMemory Lib "kernel32.dll" ( _
     ByRef Destination As Any, _
     ByRef Source As Any, _
     ByVal Length As Long)

Private boolIsDirty As Boolean

Public Enum ConversionMode
    cmCollection = 0
    cmValueArray = 1
    cm2DimKeyArray = 2
End Enum

Public RaiseInvalidKeyException As Boolean

Public Enum ConversionFlag
    Default = 0
    SetToNullIfEmpty = 1
    OnlyDirtyValues = 2
End Enum

Private m_Capacity As Long

Implements IReadOnlyCollection
Implements ICollection
Implements IDictionary
Implements ICloneable
Implements IEnumerable
Implements IDirtyKeySupport

Private Function ICloneable_Clone() As Object
    Set ICloneable_Clone = Me.Copy
End Function

Private Property Get ICollection_IsFixedSize() As Boolean
    ICollection_IsFixedSize = m_Capacity > 0
End Property

Private Property Set ICollection_Item(ByVal index As Long, ByVal RHS As Object)
    Set Me.Item(index) = RHS
End Property

Private Property Let ICollection_Item(ByVal index As Long, ByVal RHS As Variant)
    Me.Item(index) = RHS
End Property

Private Property Get ICollection_Item(ByVal index As Long) As Variant
    Call SwapVar(ICollection_Item, Me.Item(index))
End Property

Private Sub IDictionary_Add(ByVal Key As Variant, Value As Variant)
    Call Me.Add(Value, Key)
End Sub

Private Sub IDictionary_Clear()
    Call Me.Clear
End Sub

Private Function IDictionary_ContainsKey(ByVal Key As Variant) As Boolean
    IDictionary_ContainsKey = Me.Exists(Key)
End Function

'Private Function IDictionary_ContainsValue(Item As Variant) As Boolean
'    IDictionary_ContainsValue = ICollection_Contains(Item)
'End Function

Private Sub IDictionary_CopyTo(arrayTo As Variant, Optional ByVal index As Long)
    Call IReadOnlyCollection_CopyTo(arrayTo, index)
End Sub

Private Property Get IDictionary_Count() As Long
    IDictionary_Count = IReadOnlyCollection_Count
End Property

Private Function IDictionary_GetEnumerator() As IDictionaryEnumerator
    Set IDictionary_GetEnumerator = Me.GetDictionaryEnumerator
End Function

Private Property Get IDictionary_IsFixedSize() As Boolean
    IDictionary_IsFixedSize = Me.AddingAllowed = False
End Property

Private Property Get IDictionary_IsReadOnly() As Boolean
    IDictionary_IsReadOnly = ICollection_IsReadOnly
End Property

Private Property Set IDictionary_Item(ByVal index As Variant, ByVal RHS As Object)
    Set Me.Item(index) = RHS
End Property

Private Property Let IDictionary_Item(ByVal index As Variant, ByVal RHS As Variant)
    Me.Item(index) = RHS
End Property

Private Property Get IDictionary_Item(ByVal index As Variant) As Variant
    Call SwapVar(IDictionary_Item, Me.Item(index))
End Property

Private Property Get IDictionary_Keys() As IReadOnlyCollection
    Set IDictionary_Keys = Me.Keys
End Property

Private Sub IDictionary_Remove(ByVal Key As Variant)
    Call Me.Remove(Key)
End Sub

Private Function IDictionary_TryGetValue(ByVal Key As Variant, Value As Variant) As Boolean
    IDictionary_TryGetValue = TryGetValue(Key, Value)
End Function

Private Property Get IDictionary_Values() As IReadOnlyCollection
    Set IDictionary_Values = Me.Save(cmCollection)
End Property

Public Function TryGetValue(ByVal Key As Variant, ByRef Value As Variant) As Boolean
    If Me.Exists(Key) Then
        TryGetValue = True
        Value = Me.Item(Key)
    End If
End Function

Public Property Let Capacity(ByVal RHS As Long)
    If RHS < Me.Count Then
        Call Err.Raise(vbObjectError, , "New Capacity would me smaller than the hold items")
    End If

    m_Capacity = RHS
End Property

Public Property Get Capacity() As Long
    Capacity = m_Capacity
End Property

Private Sub ICollection_Add(Item As Variant)
    Call Me.Add(Item)
End Sub

'Private Property Let IReadOnlyCollection_Capacity(ByVal RHS As Long)
'    Me.Capacity = RHS
'End Property
'
'Private Property Get IReadOnlyCollection_Capacity() As Long
'    IReadOnlyCollection_Capacity = Me.Capacity
'End Property

Private Sub ICollection_Clear()
    Call Me.Clear
End Sub

Private Function ICollection_Contains(Item As Variant) As Boolean
    Dim v As Variant

    For Each v In Me
        If v = Item Then
            ICollection_Contains = True
            
            Exit Function
        End If
    Next
End Function

Friend Function RemoveByValue(Item As Variant) As Boolean
    Dim Key As Key
    
    On Error GoTo ErrHandler
    
    For Each Key In Me.Keys
        If Key.Value = Item Then
            Call Me.Remove(Key.Name)
            
            RemoveByValue = True
        
            Exit Function
        End If
    Next
    
ErrHandler:
End Function

Friend Function IndexOf(Item As Variant) As Long
    Dim Key As Key
    
    For Each Key In Me.Keys
        If Key.Value = Item Then
            IndexOf = Key.index
        
            Exit Function
        End If
    Next
    
    IndexOf = -1
End Function

Private Sub ICollection_CopyTo(arrayTo As Variant, Optional ByVal index As Long)
    Call IReadOnlyCollection_CopyTo(arrayTo, index)
End Sub

Public Sub CopyTo(arrayTo As Variant, Optional ByVal index As Long)
    Dim Temp() As Variant
    Dim i As Long
    
    Call SwapArrayPtr(Temp, Me.Save(cmValueArray))
    
    If SafeUBound(arrayTo) = -1 Then
        ReDim arrayTo(UBound(Temp))
    End If
    
    For i = 0 To UBound(Temp)
        arrayTo(i + index) = Temp(i)
    Next
End Sub

Private Sub IReadOnlyCollection_CopyTo(arrayTo As Variant, Optional ByVal index As Long)
    Call CopyTo(arrayTo, index)
End Sub

Private Property Get IReadOnlyCollection_Count() As Long
    IReadOnlyCollection_Count = Me.Count
End Property

Private Property Get ICollection_Count() As Long
    ICollection_Count = Me.Count
End Property

Private Function IReadOnlyCollection_GetEnumerator() As IEnumerator
    Set IReadOnlyCollection_GetEnumerator = Me.GetEnumerator
End Function

Private Function ICollection_GetEnumerator() As IEnumerator
    Set ICollection_GetEnumerator = Me.GetEnumerator
End Function

Private Property Get ICollection_IsReadOnly() As Boolean
    ICollection_IsReadOnly = Me.AssigningAllowed = False
End Property

Private Function ICollection_Remove(Value As Variant) As Boolean
    ICollection_Remove = RemoveByValue(Value)
End Function

Public Function GetDirtyCount() As Long
    Dim i As Long
    Dim Count As Long
    
    If boolIsDirty = True Then
        Count = Me.Count
        
        If Count > 0 Then
            For i = 0 To Count - 1
                If m_aItems(i).bDirty = True Then
                    GetDirtyCount = GetDirtyCount + 1
                End If
            Next i
        End If
    End If
End Function

Public Sub Clean()
    Dim i As Long
    
    If boolIsDirty Then
        boolIsDirty = False
        
        For i = m_nBaseIndex To Count - 1
            m_aItems(i).bDirty = False
        Next i
    End If
End Sub

Public Function Save(ByVal WrapMode As ConversionMode, Optional ByVal Flags As ConversionFlag) As Variant
    Dim var() As Variant
    Dim col As Dictionary
    Dim i As Long
    Dim lngUBound As Long
    Dim Ptr As Long
    Dim j As Long
    
    lngUBound = Count - 1
    
    Select Case WrapMode
        Case ConversionMode.cmCollection
            Set col = New Dictionary

            For i = m_nBaseIndex To lngUBound
                If ((Flags And OnlyDirtyValues) > 0 And m_aItems(i).bDirty) Or (Flags And OnlyDirtyValues) = 0 Then
                    Call col.Add(m_aItems(i).sKey, m_aItems(i).oItem)
                End If
            Next i
            
            Set Save = col
        Case ConversionMode.cmValueArray
            ReDim var(lngUBound)
        
            For i = 0 To lngUBound
                If ((Flags And OnlyDirtyValues) > 0 And m_aItems(i).bDirty) Or (Flags And OnlyDirtyValues) = 0 Then
                    If (Flags And SetToNullIfEmpty) > 0 Then
                        If IsArray(m_aItems(i).oItem) = False Then
                            If VarType(m_aItems(i).oItem) = vbBoolean Then ' v2
                                var(j) = m_aItems(i).oItem
                            ElseIf m_aItems(i).oItem = Empty Then
                                var(j) = Null
                            ElseIf VarType(m_aItems(i).oItem) = vbString Then
                                If m_aItems(i).oItem = "{00000000-0000-0000-0000-000000000000}" Then
                                    var(j) = Null
                                Else
                                    var(j) = m_aItems(i).oItem
                                End If
                            Else
                                var(j) = m_aItems(i).oItem
                            End If
                        Else
                            Call RtlMoveMemory(Ptr, ByVal VarPtr(m_aItems(i).oItem) + 8, 4)
                            
                            If Ptr > 0 Then
                                If UBound(m_aItems(i).oItem) < LBound(m_aItems(i).oItem) Then
                                    var(j) = Null
                                Else
                                    var(j) = m_aItems(i).oItem
                                End If
                            Else
                                var(j) = Null
                            End If
                        End If
                    Else
                        var(j) = m_aItems(i).oItem
                    End If
                    
                    j = j + 1
                End If
            Next i
            
            If (Flags And OnlyDirtyValues) > 0 Then
                If j < lngUBound + 1 Then
                    ReDim Preserve var(j - 1)
                End If
            End If
        
            Save = var
        Case ConversionMode.cm2DimKeyArray
            ReDim var(1, lngUBound)
            
            For i = 0 To lngUBound
                If ((Flags And OnlyDirtyValues) > 0 And m_aItems(i).bDirty) Or (Flags And OnlyDirtyValues) = 0 Then
                    var(0, j) = m_aItems(i).sKey
                    
                    If IsObject(m_aItems(i).oItem) Then
                        Set var(1, j) = m_aItems(i).oItem
                    Else
                        var(1, j) = m_aItems(i).oItem
                    End If
                    
                    j = j + 1
                End If
            Next i
            
            If (Flags And OnlyDirtyValues) > 0 Then
                If j < lngUBound + 1 Then
                    ReDim Preserve var(1, j - 1)
                End If
            End If
        
            Save = var()
        Case Else
            Error 0
    End Select
End Function

Friend Sub SetData(ByRef aKeys() As CollectionKey, nHoldKeyArrayIndex As Long, ByVal nCurrentEnumOperations As Long, _
      ByVal eCompareMode As CompareMode, ByVal bAllowItemAsignments As Boolean, ByVal nBaseIndex As Long, _
      ByRef anIndexLookp() As Long, ByRef aItems() As CollectionValueStruct)
    m_aKeys() = aKeys()
    m_nHoldKeyArrayIndex = nHoldKeyArrayIndex
    m_nCurrentEnumOperations = nCurrentEnumOperations
    m_eCompareMode = eCompareMode
    m_bAllowItemAsignments = bAllowItemAsignments
    m_nBaseIndex = nBaseIndex
    m_anIndexLookp() = anIndexLookp()
    m_aItems() = aItems()
End Sub

Public Function Copy() As InternalDictionary
    Set Copy = New InternalDictionary
    
    Call Copy.SetData( _
            m_aKeys(), _
            m_nHoldKeyArrayIndex, _
            m_nCurrentEnumOperations, _
            m_eCompareMode, _
            m_bAllowItemAsignments, _
            m_nBaseIndex, _
            m_anIndexLookp(), _
            m_aItems())
End Function

'Public Property Get Items() As CollectionValueStruct()
'    If Count > 0 Then
'        Items = m_aItems()
'    End If
'End Property

Public Function DirtyKeys() As KeyCollection
    Dim i As Long
    Dim u As Long
    
    Set DirtyKeys = New KeyCollection
    
    Dim Keys As New Collection

    If Me.Count > 0 Then
        u = UBound(m_aKeys)
        
        If u = 0 Then
            If m_aKeys(0).bInitialized = False Then
                Exit Function
            End If
        ElseIf boolIsDirty = False Then
            Exit Function
        End If
        
        For i = 0 To u
            If m_aItems(i).bDirty Then
                Call Keys.Add(m_aItems(i).sKey)
            End If
        Next i
        
        Call DirtyKeys.Init(Keys)
    End If
End Function

Public Function DirtyValues() As ValueCollection
    Dim i As Long
    Dim u As Long
    
    Set DirtyValues = New ValueCollection
    
    Dim Items As New Collection

    If Me.Count > 0 Then
        u = UBound(m_aKeys)
        
        If u = 0 Then
            If m_aKeys(0).bInitialized = False Then
                Exit Function
            End If
        ElseIf boolIsDirty = False Then
            Exit Function
        End If
        
        For i = 0 To u
            If m_aItems(i).bDirty Then
                Call Items.Add(m_aItems(i).oItem)
            End If
        Next i
        
        Call DirtyValues.Init(Items)
    End If
End Function
  
' sets the compare mode to either text or binary for keys in the collection.
' this allows you to have case sensitive or case insensitive keys
Public Property Get CompareMode() As CompareMode
  CompareMode = m_eCompareMode
End Property

Public Property Let CompareMode(ByVal eNewVal As CompareMode)
  If Me.Count > 0 Then
    ' if there are already items in the collection we can't change the case sensitivity
    Err.Raise &H80000500, , "Collection already initialized"
    
  Else
    If (eNewVal = BinaryCompare) Or (eNewVal = TextCompare) Then
      m_eCompareMode = eNewVal
    Else
      ' illeagal value
      Err.Raise 9 '<- Subscript out of range
    End If
  End If
End Property

' allows assignments to be made to the Item property
' If this is set, the user is responsible for
Public Property Get AssigningAllowed() As Boolean
    AssigningAllowed = m_bAllowItemAsignments
End Property

Public Property Get AddingAllowed() As Boolean
    AddingAllowed = Not m_bDisallowKeyCreation
End Property

Public Sub LockAsignments()
     m_bAllowItemAsignments = False
End Sub

Public Sub LockAdding()
     m_bDisallowKeyCreation = True
End Sub

' sets the number at which the collection indexes start.  this allows you
' to have zero or ANY number (including negitive numbers) as the base index
Public Property Get BaseIndex() As Long
  BaseIndex = m_nBaseIndex
End Property

Public Property Let BaseIndex(ByVal nNewVal As Long)
  If Me.Count > 0 Then
    ' if there are already items in the collection we can't change the base index
    Err.Raise &H80000500, , "Collection already initialized"
    
  Else
    If nNewVal <> m_nBaseIndex Then
      EraseArrays

      m_nBaseIndex = nNewVal
            
      InitializeArrays
    End If
  End If
End Property

Public Property Get Keys() As Keys
  ' creates a keys collection object that allows the keys to be enumerated

  Dim oKeys As Keys
  Set oKeys = New Keys

  oKeys.InitializeForEnumeration ObjPtr(Me), 0

  Set Keys = oKeys

  Set oKeys = Nothing

End Property

Friend Sub GetKeysForEnumeration(ByRef CurrentEnumRef As Long, aVariants As Variant, ByRef lRetVal As Long)
  ' callback function used to get the key info from the Keys class.  we don't do anything fancy
  ' here like we do with enumerating the collection of objects.  instead, we just pass each item
  ' and increment the counter.  if the user deletes an item, it could whack the iteration but we
  ' just aren't going to handle that for the keys.  if keeping the integrity of the keys enumeration
  ' even if deletes are made is something you need, you will need to implement something similar
  ' to the enumeration object class in the Keys class.

  ' if the current ref is within the array bounds, create, initialize and pass a new Key object.
  ' otherwise, return S_FALSE to break out of the enumeration
  If CurrentEnumRef <= UBound(m_aKeys) Then
    Dim oKey As Key
    Set oKey = New Key

    With m_aKeys(CurrentEnumRef)
      oKey.Initialize .sKey, .nIndex, m_aItems(m_anIndexLookp(.nIndex)).oItem
    End With

    Set aVariants = oKey

  Else
    lRetVal = S_FALSE
  End If

End Sub

Friend Function GetKeysCount() As Long
  ' return the count of the keys array
  If m_aKeys(0).bInitialized Then GetKeysCount = UBound(m_aKeys, 1) + 1
End Function

Friend Property Get Key(ByVal nIndex As Long) As Key
  ' returns the Key value associated with the passed index
  
  ' make sure the items are within the bounds of the array
  If (nIndex >= m_nBaseIndex) And (nIndex <= UBound(m_anIndexLookp, 1)) Then
    Set Key = New Key
    
    With m_aItems(m_anIndexLookp(nIndex))
        Call Key.Initialize(.sKey, nIndex, .oItem)
    End With
  Else
    Err.Raise 9 '<- Subscript out of range
  End If
  
End Property

Public Function index(ByVal sKey As String) As Long
  ' returns the index of an item if the key exists in the collection or KEY_NOT_FOUND
    index = FindKey(sKey)
End Function

Public Function Exists(ByVal sKey As String) As Boolean
  ' returns the index of an item if the key exists in the collection or KEY_NOT_FOUND
    Exists = FindKey(sKey) <> KEY_NOT_FOUND
End Function

Private Function FindKey(ByVal sKey As String, Optional ByRef nKeyLookupIndex As Long, Optional ByVal bAddIfNotFound As Boolean, _
                                                                                                Optional ByRef nClosestMatch As Long) As Long
  ' - searches for the desired key in the collection.
  '   returns the array index number if found or KEY_NOT_FOUND (- &HEFFFFFFF) if not found.
  '   we return KEY_NOT_FOUND instead of zero when a key is not found to allow collections to have any base index
  ' - Also returns the index in the lookup key array if nKeyLookupIndex was passed.
  ' - Additionally, if bAddIfNotFound is set, a new lookup key array item is added with the passed key.
  ' - nClosestMatch is an [out] param that holds the index of the closest match that was found in the key lookup array.

  Dim i As Long, nUbound As Long, nLow As Long, nHigh As Long

  ' assume failure
  nKeyLookupIndex = (-1)
    
  If Len(sKey) = 0 Then
    ' short circuit and return value of KEY_NOT_FOUND tells the caller no match was found
    FindKey = KEY_NOT_FOUND
  
  Else
    nUbound = UBound(m_aKeys, 1)
    
    nHigh = nUbound
    
    ' check to see if this is the first item to be added to the collection
    If (i = nHigh) And (m_aKeys(i).bInitialized = False) Then
      ' the only place the bAddIfNotFound flag is set is in the Add method.  if the flag is set
      ' an item is being added to the collection and we need to add the corresponding entry in
      ' the key lookup array.
      If bAddIfNotFound Then
        With m_aKeys(i)
          ' set the flag indicating this member has been initialized
          .bInitialized = True
          
          nUbound = UBound(m_anIndexLookp, 1)
          
          ' an item is being added so we need to add the index of the "about to be added" item to the lookup array
          If m_anIndexLookp(nUbound) <> KEY_NOT_FOUND Then
            .nIndex = nUbound + 1
          Else
            .nIndex = nUbound
          End If
          
          .sKey = sKey
        End With
      End If
      
      ' return value of KEY_NOT_FOUND tells the caller no match was found
      FindKey = KEY_NOT_FOUND
  
    Else
      
      Do
        ' divide and conquer!  Each time we loop devide the difference between the
        ' last items checked and search between the two indexes.  This is MUCH faster
        ' than looping through the entire list when dealing with a sorted array.
        i = nLow + ((nHigh - nLow) / 2)
        
        ' see how sKey relates to the current index....
        Select Case StrComp(sKey, m_aKeys(i).sKey, m_eCompareMode)
          Case 0
            FindKey = m_aKeys(i).nIndex
            
            ' return the lookup index
            nKeyLookupIndex = i
            nClosestMatch = i
            Exit Do
            
          Case (-1): nHigh = i - 1
          Case 1: nLow = i + 1
            
          Case Else
            ' raise error here
        End Select
    
        
        ' if the low search bound has become greater than the high search bound, the
        ' item does not exist in the array.  if the bAddIfNotFound flag is set, a new
        ' item is being added.  otherwise, just return the not found value.
        If nLow > nHigh Then
          If bAddIfNotFound Then
            ' check to see whether or not this item is initialized
            If m_aKeys(nUbound).bInitialized Then
              ReDim Preserve m_aKeys(nUbound + 1)
          
              nUbound = UBound(m_aKeys, 1)
            End If
            
            With m_aKeys(nUbound)
              .bInitialized = True
              
              ' bAddIfNotFound is only set from the Add method in this class.
              ' if this flag is set, a new item is about to be added to the m_anIndexLookp array
              ' So, we need to assign the index of the "about to be added" item to the lookup array member
              If m_anIndexLookp(UBound(m_anIndexLookp)) <> KEY_NOT_FOUND Then
                .nIndex = UBound(m_anIndexLookp, 1) + 1
              Else
                .nIndex = UBound(m_anIndexLookp, 1)
              End If
              
              .sKey = sKey
            End With
            
  
            ' see whether we should add this item above or below the item at index 'i'
            Select Case StrComp(sKey, m_aKeys(i).sKey, m_eCompareMode)
              
              ' return the lookup index
              Case (-1): nKeyLookupIndex = i
              Case 1: nKeyLookupIndex = i + 1
                
            End Select
            
            Call MoveKey(nUbound, nKeyLookupIndex)
            
          End If ' bAddIfNotFound
          
          ' return the item key lookup array that was the closest match
          nClosestMatch = i
          
          ' return value of KEY_NOT_FOUND tells the caller no match was found
          FindKey = KEY_NOT_FOUND
          Exit Do
        End If ' nLow > nHigh
    
      Loop
    End If ' (i = nHigh) And (m_aKeys(i).bInitialized = False) ... Else
  End If ' sKey = vbnullstring
  
End Function

Private Function MoveKey(ByVal nMovedKey As Long, ByVal nTargetPosition As Long) As Long
  ' moves an item in the key lookup array to a new index
  
  Dim i As Long, nLbound As Long, nUbound As Long, nStructLen As Long, abytBuffer() As Byte
  
  If nMovedKey = nTargetPosition Then
    ' don't need to move item
    MoveKey = nTargetPosition
  
  Else
    nStructLen = LenB(m_aKeys(nMovedKey))
  
    ReDim abytBuffer(nStructLen - 1)
  
    nUbound = UBound(m_aKeys, 1)
    
    ' make sure the target position isn't greater than the ubound of the array
    If nTargetPosition > nUbound Then nTargetPosition = nUbound
  
    ' copy the item to be moved
    CopyMemory ByVal VarPtr(abytBuffer(0)), ByVal VarPtr(m_aKeys(nMovedKey)), nStructLen
  
    Select Case True
      Case nMovedKey < nTargetPosition
        ' grab all of the items below the moved one and shift them up
        CopyMemory ByVal VarPtr(m_aKeys(nMovedKey)), ByVal VarPtr(m_aKeys(nMovedKey + 1)), (nTargetPosition - nMovedKey) * nStructLen
    
        ' place the moved item into the position that used to be held by the target
        CopyMemory ByVal VarPtr(m_aKeys(nTargetPosition)), ByVal VarPtr(abytBuffer(0)), nStructLen
        
    
      Case nMovedKey > nTargetPosition
        ' grab all of the items above the moved one and shift them down
        CopyMemory ByVal VarPtr(m_aKeys(nTargetPosition + 1)), ByVal VarPtr(m_aKeys(nTargetPosition)), (nMovedKey - nTargetPosition) * nStructLen
    
        ' place the moved item into the position that used to be held by the target
        CopyMemory ByVal VarPtr(m_aKeys(nTargetPosition)), ByVal VarPtr(abytBuffer(0)), nStructLen
              
    End Select
    
    MoveKey = nTargetPosition
  End If
  
End Function

Public Sub ChangeKey(ByVal sOldKey$, ByVal sNewKey$)
  ' allow for key changes
  
  Dim nItem As Long, nClosestKeyLookupIndex As Long, nKeyLookupIndex As Long, nUbound As Long
  
  ' first, make sure the new key isn't already in the collection
  If FindKey(sNewKey, , , nClosestKeyLookupIndex) <> KEY_NOT_FOUND Then
    Err.Raise 457 '<- This key is already associated with an element of this collection

  Else
    ' find the index of the item with the given key.
    ' this also returns the index into the lookup item array for the key
    nItem = FindKey(sOldKey, nKeyLookupIndex)
  
    If nItem <> KEY_NOT_FOUND Then
      With m_aItems(m_anIndexLookp(nItem))
        .sKey = sNewKey
        
        ' the one place where we manipulate the key in the child object
        .oItem.Key = sNewKey
      End With
      
      SetChangedKeyPosition nKeyLookupIndex, sNewKey, nClosestKeyLookupIndex, nUbound, nItem
      
    Else
      Err.Raise KeyNotFound, , "Key not found."
    End If ' nItem <> KEY_NOT_FOUND
  End If ' FindKey(sNewKey) <> KEY_NOT_FOUND
  
End Sub

Public Sub ChangeKeyForIndex(ByVal nIndex$, ByVal sNewKey$)
  ' allow for key changes
  
  Dim i As Long, nClosestKeyLookupIndex As Long, nKeyLookupIndex As Long, nUbound As Long
  
  If (nIndex >= m_nBaseIndex) And (nIndex <= UBound(m_anIndexLookp, 1)) Then
    ' make sure the new key isn't already in the collection
    If FindKey(sNewKey, , , nClosestKeyLookupIndex) <> KEY_NOT_FOUND Then
      Err.Raise 457 '<- This key is already associated with an element of this collection
    
    Else
      With m_aItems(m_anIndexLookp(nIndex))
        .sKey = sNewKey
        
        ' the one place where we manipulate the key in the child object
        .oItem.Key = sNewKey
      End With
      
      nUbound = UBound(m_aKeys, 1)
      
      nKeyLookupIndex = (-1)
      
      ' find the item in the lookup array
      For i = 0 To nUbound
        If m_aKeys(i).nIndex = nIndex Then
          nKeyLookupIndex = i
          Exit For
        End If
      Next
      
      
      SetChangedKeyPosition nKeyLookupIndex, sNewKey, nClosestKeyLookupIndex, nUbound, nIndex
      
      
    End If ' FindKey(sNewKey, , , nClosestKeyLookupIndex) <> KEY_NOT_FOUND
    
  Else ' (nIndex >= m_nBaseIndex) And (nIndex <= UBound(m_anIndexLookp))
    ' raise error here
  End If ' (nIndex >= m_nBaseIndex) And (nIndex <= UBound(m_anIndexLookp))
  
End Sub

Private Sub SetChangedKeyPosition(ByVal nKeyLookupIndex As Long, ByVal sNewKey$, ByVal nClosestKeyLookupIndex As Long, ByVal nUbound As Long, ByVal nIndex As Long)
  ' handles moving and setting the values for the item in the key lookup array for both ChangKey subs
  
  If (nKeyLookupIndex <> (-1)) And (sNewKey <> vbNullString) Then
    m_aKeys(nKeyLookupIndex).sKey = sNewKey
    
    ' see whether we should add this item above or below the item at index 'nClosestKeyLookupIndex'
    Select Case StrComp(sNewKey, m_aKeys(nClosestKeyLookupIndex).sKey, m_eCompareMode)
      Case (-1): nClosestKeyLookupIndex = nClosestKeyLookupIndex '- 1
      Case 1: nClosestKeyLookupIndex = nClosestKeyLookupIndex + 1
    End Select
    
    nUbound = UBound(m_aKeys, 1)

    If nClosestKeyLookupIndex > (nUbound + 1) Then nClosestKeyLookupIndex = nUbound + 1
    
    Call MoveKey(nKeyLookupIndex, nClosestKeyLookupIndex)
  
  
  Else
  
    If nKeyLookupIndex <> (-1) Then
      ' now remove the item from the key lookup array....
      Call MoveKey(nKeyLookupIndex, nUbound)
    
      ' if the ubound is greater than zero, preserve the data....
      If nUbound Then
        ReDim Preserve m_aKeys(nUbound - 1)
      Else
        ReDim m_aKeys(0)
      End If
    End If
    
    
    ' add the new key to the lookup array
    If sNewKey <> vbNullString Then
      
      ' check to see whether or not this item is initialized
      If m_aKeys(nUbound).bInitialized Then
        ReDim Preserve m_aKeys(nUbound + 1)
    
        nUbound = UBound(m_aKeys, 1)
      End If
      
      With m_aKeys(nUbound)
        .bInitialized = True
        .nIndex = nIndex
        .sKey = sNewKey
      End With
      
      ' see whether we should add this item above or below the item at index 'nClosestKeyLookupIndex'
      Select Case StrComp(sNewKey, m_aKeys(nClosestKeyLookupIndex).sKey, m_eCompareMode)
        Case (-1): nKeyLookupIndex = nClosestKeyLookupIndex '- 1
        Case 1: nKeyLookupIndex = nClosestKeyLookupIndex + 1
      End Select
      
      If nClosestKeyLookupIndex > UBound(m_aKeys, 1) Then nClosestKeyLookupIndex = UBound(m_aKeys, 1)
     
      Call MoveKey(nUbound, nKeyLookupIndex)
      
    End If ' sNewKey <> vbnullstring
    
  End If ' (nKeyLookupIndex <> (-1)) And (sNewKey <> vbnullstring)

End Sub

Public Sub Add( _
        vItem As Variant, _
        Optional ByVal sKey As String, _
        Optional ByVal eInsert As CollectionAddOption = EndOfCollection, _
        Optional ByVal nPosition As Long = KEY_NOT_FOUND)
  
  Dim nLbound As Long, nUbound As Long, nKeyArrayIndex As Long, nFindKey As Long
  Dim Ptr As Long
  
'    If Len(sKey) = 0 Then
'        Error 0
'    End If
  
  nLbound = m_nBaseIndex
  nUbound = UBound(m_anIndexLookp, 1)
  
    If m_bDisallowKeyCreation Then
        Error 0
    End If
  
  ' make sure the specified position is within the bounds of the array
  If nPosition <> KEY_NOT_FOUND Then
    Select Case nPosition
      Case Is < nLbound, Is > nUbound
        Err.Raise 9 '<- Subscript out of range
        
        Exit Sub
        
      Case nLbound
        If eInsert = AddBeforePosition Then eInsert = BeginningOfCollection
        
      Case nUbound
        If eInsert = AddAfterPosition Then eInsert = EndOfCollection
        
    End Select
    
  ElseIf ((eInsert = AddBeforePosition) Or (eInsert = AddAfterPosition)) And (nPosition = KEY_NOT_FOUND) Then
    Err.Raise NoPositionSupplied, , "No position supplied."
    
    Exit Sub
  End If
  
  If sKey <> vbNullString Then
    If Len(sKey) = 0 Then
        nFindKey = FindKey(CStr(StrPtr(sKey)), nKeyArrayIndex, True)
    Else
        nFindKey = FindKey(sKey, nKeyArrayIndex, True)
    End If
  ElseIf IsObject(vItem) Then
    nFindKey = FindKey(CStr(ObjPtr(vItem)), nKeyArrayIndex, True)
Else
    nFindKey = KEY_NOT_FOUND
  End If
  
  ' first thing to do is make sure the key does not already exist in the collection
  If nFindKey <> KEY_NOT_FOUND Then
    Err.Raise 457 '<- This key is already associated with an element of this collection
    
  Else
    ' check to see whether or not this item is initialized
    If m_anIndexLookp(nUbound) <> KEY_NOT_FOUND Then
      ReDim Preserve m_anIndexLookp(m_nBaseIndex To nUbound + 1)
      
      If m_aItems(UBound(m_aItems, 1)).bInitialized Then
        ReDim Preserve m_aItems(UBound(m_aItems, 1) + 1)
      End If
      
      nUbound = UBound(m_anIndexLookp, 1)
    End If
    
    m_anIndexLookp(nUbound) = UBound(m_aItems, 1)
    
    ' set the object and the key into the array
    With m_aItems(m_anIndexLookp(nUbound))
        .bInitialized = True
      
        If IsObject(vItem) Then
            Set .oItem = vItem
            
            If Len(sKey) = 0 Then
                .sKey = CStr(ObjPtr(vItem))
            Else
                .sKey = sKey
            End If
        ElseIf VarType(vItem) = vbString And Len(sKey) = 0 And StrPtr(sKey) > 0 Then
            .oItem = vItem
            .sKey = CStr(StrPtr(sKey))
        Else
            .oItem = vItem
            .sKey = sKey
            
            If IsArray(.oItem) = False Then
                If VarType(vItem) <> vbUserDefinedType Then
                    .bDirty = vItem <> Empty
                Else
                    .bDirty = True
                End If
            Else
                Call RtlMoveMemory(Ptr, ByVal VarPtr(vItem) + 8, 4)
                
                If (Ptr = 0) = False Then
                    If UBound(.oItem) > LBound(.oItem) Then
                        .bDirty = True
                    End If
                End If
            End If
        End If
    End With
        
    
    ' set module level flag to avoid lookup hit in the MoveItem method
    m_nHoldKeyArrayIndex = nKeyArrayIndex
    
    ' if we were directed to add the item before or after another item, move it to the desired location
    Select Case eInsert
      Case BeginningOfCollection
        Me.MoveItem nUbound, nLbound
        
      Case AddBeforePosition
        Me.MoveItem nUbound, nPosition
      
      Case AddAfterPosition
        Me.MoveItem nUbound, nPosition + 1
    
    End Select
    
  End If

End Sub

Public Property Get Count() As Long
    On Error GoTo Fehler

  ' count of the objects in the collection.
  If m_anIndexLookp(UBound(m_anIndexLookp, 1)) <> KEY_NOT_FOUND Then
    Count = (UBound(m_anIndexLookp, 1) + 1) - m_nBaseIndex
  End If
  
  Exit Property
Fehler:
End Property

Public Sub RemoveAll()
    Dim i As Long
    Dim Key As Key
    Dim oKeys As Keys
    Dim boolKeyNotFound As Boolean
    Dim lngCount As Long
    
'    Debug.Assert 0

    If Not m_bAllowItemAsignments Then
        Call Err.Raise(vbObjectError + 100, "RemoveAll", "Removing is not allowed then collection is locked.")
    
        Exit Sub
    End If
    
    Set oKeys = New Keys

    Call oKeys.InitializeForEnumeration(ObjPtr(Me), 0)
    
    For Each Key In oKeys
        If Len(Key.Name) > 0 Then
            Call Me.Remove(Key.Name)
        Else
            boolKeyNotFound = True
        End If
    Next Key
    
    If boolKeyNotFound Then
        If m_anIndexLookp(UBound(m_anIndexLookp, 1)) <> KEY_NOT_FOUND Then
          lngCount = (UBound(m_anIndexLookp, 1) + 1) - m_nBaseIndex
        End If
        
        For i = lngCount To 1 Step -1
            Call Me.Remove(i)
        Next i
    End If
    
    Call InitializeArrays
    
    Set oKeys = Nothing
End Sub

Public Sub Remove(vIndexKey As Variant)
  ' removes the item at the specified index or associated
  ' with the specified key from the collection
  
  If Not m_bAllowItemAsignments Then
        Call Err.Raise(vbObjectError + 100, "RemoveAll", "Removing is not allowed then collection is locked.")
    
        Exit Sub
    End If
  
  Dim nUbound As Long, nItem As Long, nKeyLookupIndex As Long, bFound As Boolean
  
  ' determine whether this is an index or a string key
  Select Case VarType(vIndexKey)
    Case vbSingle, vbDouble, vbByte, vbLong, vbInteger
      nItem = vIndexKey
      
      If (nItem >= m_nBaseIndex) And (nItem <= UBound(m_anIndexLookp, 1)) Then
        bFound = True
        Call FindKey(m_aItems(m_anIndexLookp(nItem)).sKey, nKeyLookupIndex)
      Else
        Err.Raise 9 '<- Subscript out of range
      End If
      
    ' if it's a string, look for it in the array
    Case vbString
      nItem = FindKey(CStr(vIndexKey), nKeyLookupIndex)
      
      If nItem <> KEY_NOT_FOUND Then bFound = True
    Case Else
        If IsObject(vIndexKey) Then
         nItem = FindKey(CStr(ObjPtr(vIndexKey)), nKeyLookupIndex)
    
        If nItem <> KEY_NOT_FOUND Then bFound = True
        Else
      Err.Raise InvalidDataType, , "Invalid datatype"
      End If
  End Select
  
  
  If bFound Then
    ' if we found the item, move it to the end of the array.
    nUbound = MoveItemToEndOfArray(nItem)
    
    ' destroy the reference
    With m_aItems(m_anIndexLookp(nUbound))
      If IsObject(.oItem) Then
        Set .oItem = Nothing
      Else
        .oItem = 0
      End If
      
      .bInitialized = False
      .sKey = vbNullString
    End With
    
    ' if the ubound is greater than the collection base number, preserve the data....
    If nUbound > m_nBaseIndex Then
      ReDim Preserve m_anIndexLookp(m_nBaseIndex To nUbound - 1)
    Else
      ReDim m_anIndexLookp(m_nBaseIndex To m_nBaseIndex)
      m_anIndexLookp(m_nBaseIndex) = KEY_NOT_FOUND
    End If
  
    ' now remove the key lookup array item
    If nKeyLookupIndex > (-1) Then
      nUbound = UBound(m_aKeys, 1)
    
      nItem = MoveKey(nKeyLookupIndex, nUbound)
    
      ' if the ubound is greater than zero, preserve the data....
      If nUbound Then
        ReDim Preserve m_aKeys(nUbound - 1)
      Else
        ReDim m_aKeys(0)
      End If
    End If
  Else
    Err.Raise KeyNotFound, "CSCollection", "Key not found."
  End If
  
End Sub

Public Function IsDirty(Optional ByVal vIndexKey As Variant) As Boolean
  ' retrieves the specified item from the collection.
  ' this is the default property of the collection interface
  ' which allows us to access an item via the following syntax:
  '
  '  Dim oObject As cCollectionChildObject
  '
  '  Set oObject = oCollection('index or Key')
  '
  
  Dim nItem As Long, bFound As Boolean
    
    If IsMissing(vIndexKey) Then
        IsDirty = boolIsDirty
    Else
      ' determine whether this is an index or a string key
          Select Case VarType(vIndexKey)
    Case vbSingle, vbDouble, vbByte, vbLong, vbInteger
      nItem = vIndexKey
      
      If (nItem >= m_nBaseIndex) And (nItem <= UBound(m_anIndexLookp)) Then
        If m_anIndexLookp(nItem) <> KEY_NOT_FOUND Then bFound = True
      End If
      
    ' if it's a string, look for it in the array
    Case vbString
      nItem = FindKey(CStr(vIndexKey))
      
      If nItem <> KEY_NOT_FOUND Then bFound = True
    Case Else
    If IsObject(vIndexKey) Then
        nItem = FindKey(CStr(ObjPtr(vIndexKey)))
      
      If nItem <> KEY_NOT_FOUND Then bFound = True
    Else
      Err.Raise InvalidDataType, , "Invalid datatype"
End If
  End Select

  If bFound Then
    IsDirty = m_aItems(m_anIndexLookp(nItem)).bDirty
    'Set Item = m_aItems(m_anIndexLookp(nItem)).oItem
  Else
'    If RaiseInvalidKeyException Then
        Err.Raise KeyNotFound, , "Key not found."
'    End If
  End If
    End If
  
End Function

Public Property Get Item(vIndexKey As Variant) As Variant
Attribute Item.VB_UserMemId = 0
  ' retrieves the specified item from the collection.
  ' this is the default property of the collection interface
  ' which allows us to access an item via the following syntax:
  '
  '  Dim oObject As cCollectionChildObject
  '
  '  Set oObject = oCollection('index or Key')
  '
  
  Dim nItem As Long, bFound As Boolean

  ' determine whether this is an index or a string key
  Select Case VarType(vIndexKey)
    Case vbSingle, vbDouble, vbByte, vbLong, vbInteger
      nItem = vIndexKey
      
      If (nItem >= m_nBaseIndex) And (nItem <= UBound(m_anIndexLookp)) Then
        If m_anIndexLookp(nItem) <> KEY_NOT_FOUND Then bFound = True
      End If
      
    ' if it's a string, look for it in the array
    Case vbString
      nItem = FindKey(CStr(vIndexKey))
      
      If nItem <> KEY_NOT_FOUND Then bFound = True
    Case Else
    If IsObject(vIndexKey) Then
        nItem = FindKey(CStr(ObjPtr(vIndexKey)))
      
      If nItem <> KEY_NOT_FOUND Then bFound = True
    Else
      Err.Raise InvalidDataType, , "Invalid datatype"
End If
  End Select

  If bFound Then
    With m_aItems(m_anIndexLookp(nItem))
      If IsObject(.oItem) Then
        Set Item = .oItem
      Else
        Item = .oItem
      End If
    End With
    'Set Item = m_aItems(m_anIndexLookp(nItem)).oItem
  Else
    If RaiseInvalidKeyException Then
        Err.Raise KeyNotFound, , "Key not found."
    End If
  End If
  
End Property

Public Property Let Item(vIndexKey As Variant, vNewItem As Variant)
    Dim Ptr As Long
    Dim Ptr2 As Long

  If m_bAllowItemAsignments Then
  
    Dim nItem As Long, bFound As Boolean
    
    If IsObject(vIndexKey) Then
         If vIndexKey Is Nothing Then
            Err.Raise InvalidDataType, , "Invalid datatype (Nothing)"
        Else
            nItem = FindKey(CStr(ObjPtr(vIndexKey)))
    
            If nItem <> KEY_NOT_FOUND Then bFound = True
        End If
    Else
        ' determine whether this is an index or a string key
        Select Case VarType(vIndexKey)
          Case vbSingle, vbDouble, vbByte, vbLong, vbInteger
            nItem = vIndexKey
    
            If (nItem >= m_nBaseIndex) And (nItem <= UBound(m_anIndexLookp, 1)) Then bFound = True
    
          ' if it's a string, look for it in the array
          Case vbString
            nItem = FindKey(CStr(vIndexKey))
    
            If nItem <> KEY_NOT_FOUND Then bFound = True
          Case Else
            Err.Raise InvalidDataType, , "Invalid datatype"
        End Select
    End If

    If Not bFound Then
        If m_bDisallowKeyCreation Then
            Call Err.Raise(vbObjectError, "Item", "Key creation disallowed.")
            'Error 0
        End If
    
        Call Add(vNewItem, vIndexKey)
        
       Exit Property
    End If

    If bFound Then
'      ' this is the one place where we diverge from a generic encapsulated object model.
'      ' we need to check the key property of the new object to see whether or not we need to
'      ' change our internal key.  if the keys are not the same, change our copy.
'      If Not vNewItem Is Nothing Then
'        If StrComp(vNewItem.Key, m_aItems(nItem).sKey, m_eCompareMode) Then
'          Me.ChangeKeyForIndex nItem, vNewItem.Key
'        End If
'      End If
      
      With m_aItems(m_anIndexLookp(nItem))
        If IsObject(.oItem) Then
            .bDirty = (.oItem Is vNewItem = False) And bFound = True
            Set .oItem = vNewItem
        ElseIf bFound = True Then
            If IsArray(.oItem) = False Then
                If VarType(.oItem) <> vbUserDefinedType Then
                    If .oItem <> vNewItem Then
                        .bDirty = True
                    End If
                Else
                    .bDirty = True
                End If
                
                If .bDirty Then
                    .oItem = vNewItem
                End If
            Else
                Call RtlMoveMemory(Ptr, ByVal VarPtr(.oItem) + 8, 4)
                Call RtlMoveMemory(Ptr2, ByVal VarPtr(vNewItem) + 8, 4)
                
                If ((Ptr2 = 0) And (Ptr = 0)) = False Then
                    .bDirty = True
                    .oItem = vNewItem
                End If
            End If
        Else
            Debug.Assert 0
            .oItem = vNewItem
            
            If IsArray(.oItem) = False Then
                .bDirty = vNewItem <> Empty
            Else
                Call RtlMoveMemory(Ptr, ByVal VarPtr(vNewItem) + 8, 4)
                
                If (Ptr = 0) = False Then
                    .bDirty = True
                End If
            End If
        End If
        
        If .bDirty = True Then
            boolIsDirty = True
        End If
      End With
      'Set m_aItems(m_anIndexLookp(nItem)).oItem = vNewItem
      
    Else
      Err.Raise KeyNotFound, , "Key not found."
    End If
  
  Else
    ' not allowint item assignments.  raise the same error as the VB collection
    Err.Raise 438
  End If
  
End Property

Public Property Set Item(vIndexKey As Variant, vNewItem As Variant)
    Item(vIndexKey) = vNewItem
End Property

Public Function ContainsValue(Value As Variant) As Boolean
    Dim Item As Variant
    
    For Each Item In Me
        If Item = Value Then
            ContainsValue = True
            
            Exit Function
        End If
    Next
End Function

' #############  Begin Required for enumeration block  ###################
Public Function GetDictionaryEnumerator() As IDictionaryEnumerator
  ' this property allows you to enumerate the collection with the For...Each syntax

  Dim i As Long, nBase As Long, nUbound As Long, oEnumerator As Enumerator
  
  Set oEnumerator = New Enumerator
  
  nUbound = UBound(m_anIndexLookp, 1)
  
  If Me.Count > 0 Then
    nBase = m_nBaseIndex
  Else
    nBase = nUbound + 1
  End If
  
  oEnumerator.UseDictionaryEntry = True
    
  oEnumerator.InitializeEnumeration ObjPtr(Me), nBase, nBase, nUbound, VarPtr(m_anIndexLookp(m_nBaseIndex))
  
  
  m_nCurrentEnumOperations = m_nCurrentEnumOperations + 1
  
  ' return the enumerator object's IEnumVARIANTReDef interface
  Set GetDictionaryEnumerator = oEnumerator
  
  ' clean up
  Set oEnumerator = Nothing
  
End Function

Public Function GetEnumerator() As IEnumerator
  ' this property allows you to enumerate the collection with the For...Each syntax

  Dim i As Long, nBase As Long, nUbound As Long, oEnumerator As Enumerator
  
  Set oEnumerator = New Enumerator
  
  nUbound = UBound(m_anIndexLookp, 1)
  
  If Me.Count > 0 Then
    nBase = m_nBaseIndex
  Else
    nBase = nUbound + 1
  End If

  oEnumerator.InitializeEnumeration ObjPtr(Me), nBase, nBase, nUbound, VarPtr(m_anIndexLookp(m_nBaseIndex))
  
  m_nCurrentEnumOperations = m_nCurrentEnumOperations + 1
  
  ' return the enumerator object's IEnumVARIANTReDef interface
  Set GetEnumerator = oEnumerator
  
  ' clean up
  Set oEnumerator = Nothing
  
End Function

Friend Sub GetItemsForIEnumVARIANT(ByRef CurrentEnumRef As Long, anLookup() As Long, Entry As DictionaryEntry, ByRef lRetVal As Long)
  ' this method is called from the function IEnumVARIANT_Next in the MIEnumVARIANT bas module.
  ' it is an addition to the IEnumVARIANT interface provided in the 'ienumv.tlb' typelib.  this
  ' method provides a backdoor into this class via the 'this' pointer passed to IEnumVARIANT_Next by COM.
  ' in this way, the data in this class can be accessed from the IEnumVARIANT_Next function without jumping
  ' through hoops and creating circular references.  only the aVariants and lRetVal parameters are used in
  ' this implimentation.

  Dim i As Long, j As Long, nUbound As Long, nUboundLookup As Long, nDifference As Long, bResetArray As Boolean, bExitDo As Boolean
  
  ' assume no item will be returned
  lRetVal = S_FALSE
  
  nUboundLookup = UBound(anLookup, 1)
  
  ' if the counter is greater than the upper bounds of the
  ' array, return S_FALSE ('1') to break out of the enumeration
  If CurrentEnumRef <= nUboundLookup Then
    
    ' see if there have been any changes to the array in the parent collection while we were gone....
    nUbound = UBound(m_anIndexLookp, 1)
      
    ' what we are concerened with is items that may have been added to or removed from the collection.
    ' if the array upper bound in the enumerator object is less than that of the collection, we need to
    ' investigate further.  if the upper bounds of the array in the enumerator object is greater than
    ' that of the collection, it would indicate that an item had been removed from the collection.  we
    ' handle this in the "Do" loop later in this routine....
    If nUboundLookup < nUbound Then
      ' see if the items pointed to by the array match.  if they do, it would indicate that whatever happened
      ' to the array was above the current index and should thus be reflected in the enumerator object.  we
      ' will redimention the array in the enumerator object and copy the current array from the collection into it
      If m_anIndexLookp(CurrentEnumRef) <> anLookup(CurrentEnumRef) Then
      
        ' an item may have been added or removed below this item.  let's look forward to see if we can find our item.
        ' if we find it, set the pointer to the index at which the item was found set the flag to reset the array.
        ' if we don't find the item, a move or other complex operation could have been preformed in which case, we
        ' will just proceed since we could look forever trying to figure out what had happened.   there is a count of
        ' current enumeration operations held in the m_nCurrentEnumOperations variable.  at this writing, this var is
        ' checked in the MoveItem method of this class and no moves are allowed while enumeration operations are in
        ' progress.
          
          nDifference = nUbound = nUboundLookup
          
          For j = CurrentEnumRef To CurrentEnumRef + nDifference
            If m_anIndexLookp(j) = anLookup(CurrentEnumRef) Then
              CurrentEnumRef = j
              bResetArray = True
              Exit For
            End If
          Next
        
      Else
        ' the indexes match so we will just reset the array
        bResetArray = True
      End If
      
      If bResetArray Then
        ' copy the new array data into the array in the calling object
        ReDim anLookup(m_nBaseIndex To nUbound)
        
        CopyMemory ByVal VarPtr(anLookup(m_nBaseIndex)), ByVal VarPtr(m_anIndexLookp(m_nBaseIndex)), ((nUbound - m_nBaseIndex) + 1) * 4&
      End If
      
    End If
    
    ' here we look to see if the item pointed to by the lookup array in the enumerator object is still
    ' initialized.  if it isn't, we loop through the lookup array until we find an item that is initialized
    ' or we get to the end of the the lookup array.  we also change the CurrentEnumRef pointer as needed
    i = CurrentEnumRef
    
    Do
      With m_aItems(anLookup(i))
        If .bInitialized Then
          If IsObject(.oItem) Then
            Set Entry.Value = .oItem
          Else
            Entry.Value = .oItem
          End If
          
          Entry.Key = .sKey
          
          If i <> CurrentEnumRef Then
            CurrentEnumRef = CurrentEnumRef + (i - CurrentEnumRef)
          End If
          
          lRetVal = S_OK
          bExitDo = True
        End If
      End With
      
      If bExitDo Then Exit Do
      
      i = i + 1
    Loop Until i > nUboundLookup
  End If
    
End Sub

Friend Sub DecrementEnumCounter()
  m_nCurrentEnumOperations = m_nCurrentEnumOperations - 1
End Sub
' #############  End Required for enumeration block  ###################

Private Sub Class_Initialize()
  
  MIEnumVARIANT.AddPointerToLookupList ObjPtr(Me)
  
  ' set the default collection base
  m_nBaseIndex = 1
  
  ' call the sub to initialize the arrays
  InitializeArrays
  
  RaiseInvalidKeyException = True
  m_bAllowItemAsignments = True
End Sub

Private Sub Class_Terminate()
  ' terminate the class by looping through the array and setting each
  ' object to Nothing and releasing the memory held by the arrays
  
  MIEnumVARIANT.RemovePointerFromLookupList ObjPtr(Me)
  
  ' call the sub that erases the arrays
  EraseArrays
End Sub

Private Sub InitializeArrays()
  ' initialize the arrays that will hold the collection and the key lookups
  ReDim m_anIndexLookp(m_nBaseIndex To m_nBaseIndex)
  m_anIndexLookp(m_nBaseIndex) = KEY_NOT_FOUND
  
  ReDim m_aItems(0)
  ReDim m_aKeys(0)
End Sub

Private Sub EraseArrays()
  
  Dim i As Long
  
  ReDim m_anIndexLookp(0)
  Erase m_anIndexLookp
  
  ' loop through the array data
  ' and set each item to nothing
  For i = 0 To UBound(m_aItems, 1)
    With m_aItems(i)
      If IsObject(.oItem) Then Set .oItem = Nothing
    End With
    'Set m_aItems(i).oItem = Nothing
  Next
  
  ReDim m_aItems(0)
  Erase m_aItems
  
  ReDim m_aKeys(0)
  Erase m_aKeys
  
End Sub

Public Function Clear()
  ' clear out the arrays and then reinitialize them
  EraseArrays
  InitializeArrays
End Function

Public Function MoveItem(ByVal vMovedItem As Variant, ByVal vTargetPosition As Variant) As Long
  ' moves an item in the collection to a new index in response to the user rearranging items
  
  Dim nMovedItem As Long, nTargetPosition As Long, nLbound As Long, nUbound As Long
  Dim i As Long, nKeyArrayIndex As Long, nChangedCount As Long, nChangeNum As Long, nIndex As Long
  Dim bNoMove As Boolean, nHoldForMove As Long
  
  ' only allow moves if we are not currently in any enumeration operations
  If m_nCurrentEnumOperations = 0 Then
    nKeyArrayIndex = KEY_NOT_FOUND
    
    ' determine whether this is an index or a string key for each param....
    If IsObject(vMovedItem) Then
    nMovedItem = FindKey(CStr(ObjPtr(vMovedItem)), nKeyArrayIndex)
        
        If nMovedItem = KEY_NOT_FOUND Then
          Err.Raise KeyNotFound, , "Key not found."
        Else
          ' will only hit here if we are moving an item at the users request.
          ' this is because all of the internal calls to the MoveItem method use the array index.
          m_nHoldKeyArrayIndex = nKeyArrayIndex
        End If
    Else
    Select Case VarType(vMovedItem)
      Case vbSingle, vbDouble, vbByte, vbLong, vbInteger
        nMovedItem = vMovedItem
        m_nHoldKeyArrayIndex = KEY_NOT_FOUND
        
      ' if it's a string, look for it in the array
      Case vbString
        nMovedItem = FindKey(CStr(vMovedItem), nKeyArrayIndex)
        
        If nMovedItem = KEY_NOT_FOUND Then
          Err.Raise KeyNotFound, , "Key not found."
        Else
          ' will only hit here if we are moving an item at the users request.
          ' this is because all of the internal calls to the MoveItem method use the array index.
          m_nHoldKeyArrayIndex = nKeyArrayIndex
        End If
      Case Else
        Err.Raise InvalidDataType, , "Invalid datatype"
        
    End Select
    End If
    If IsObject(vTargetPosition) Then
         nTargetPosition = FindKey(CStr(ObjPtr(vTargetPosition)))
        
        If nTargetPosition = KEY_NOT_FOUND Then
          Err.Raise KeyNotFound, , "Key not found."
        End If
        Else
    Select Case VarType(vTargetPosition)
      Case vbSingle, vbDouble, vbByte, vbLong, vbInteger
        nTargetPosition = vTargetPosition
        
      ' if it's a string, look for it in the array
      Case vbString
        nTargetPosition = FindKey(CStr(vTargetPosition))
        
        If nTargetPosition = KEY_NOT_FOUND Then
          Err.Raise KeyNotFound, , "Key not found."
        End If
      Case Else
        Err.Raise InvalidDataType, , "Invalid datatype"
    End Select
    End If
    ' .....
    
    
    nLbound = m_nBaseIndex
    nUbound = UBound(m_anIndexLookp, 1)
    
    ' make sure the items are within the bounds of the array
    Select Case True
      Case (nMovedItem > nUbound), (nTargetPosition > nUbound), (nMovedItem < nLbound), (nTargetPosition < nLbound)
        bNoMove = True
      
        Err.Raise 9 '<- Subscript out of range
        
    
      Case nMovedItem = nTargetPosition
        ' no move necessary.  return the moved item index
        bNoMove = True
        
        MoveItem = nMovedItem
  
    End Select
  
    If bNoMove = False Then
      nUbound = UBound(m_anIndexLookp, 1)
    
      ' we need to know the index of the item in the keys array.  if a lookup was called
      ' already in the chain of calls leading to this point, the index will be cached in the
      ' module level variable.
      If m_nHoldKeyArrayIndex = KEY_NOT_FOUND Then
        Call FindKey(m_aItems(m_anIndexLookp(nMovedItem)).sKey, nKeyArrayIndex)
        m_nHoldKeyArrayIndex = nKeyArrayIndex
      End If
    
    
      ' copy the item to be moved
      nHoldForMove = m_anIndexLookp(nMovedItem)
    
      Select Case True
        Case nMovedItem < nTargetPosition
          ' grab all of the items below the moved one and shift them up
          Call CopyMemory(ByVal VarPtr(m_anIndexLookp(nMovedItem)), ByVal VarPtr(m_anIndexLookp(nMovedItem + 1)), (nTargetPosition - nMovedItem) * 4&)
      
          ' place the moved item into the position that used to be held by the target
          m_anIndexLookp(nTargetPosition) = nHoldForMove
          
          
          ' now that we have moved the items, we need to fix the indexes held in the
          ' keys array.  we try to limit our time looping by keeping track of the
          ' number of indexes we *fix*.  once we have fixed all the indexes that need to
          ' be fixed, we don't need to loop any more.
          nChangeNum = nTargetPosition - nMovedItem
          
          For i = 0 To UBound(m_aKeys, 1)
            nIndex = m_aKeys(i).nIndex
  
            If (nIndex > nMovedItem) And (nIndex <= nTargetPosition) Then
              m_aKeys(i).nIndex = nIndex - 1
              
              nChangedCount = nChangedCount + 1
              
              If nChangedCount >= nChangeNum Then Exit For
            End If
          Next
      
      
        Case nMovedItem > nTargetPosition
          ' grab all of the items above the moved one and shift them down
          Call CopyMemory(ByVal VarPtr(m_anIndexLookp(nTargetPosition + 1)), ByVal VarPtr(m_anIndexLookp(nTargetPosition)), (nMovedItem - nTargetPosition) * 4&)
      
          ' place the moved item into the position that used to be held by the target
          m_anIndexLookp(nTargetPosition) = nHoldForMove
          
          ' now that we have moved the items, we need to fix the indexes held in the
          ' keys array.  we try to limit our time looping by keeping track of the
          ' number of indexes we *fix*.  once we have fixed all the indexes that need to
          ' be fixed, we don't need to loop any more.
          nChangeNum = nMovedItem - nTargetPosition
          
          For i = 0 To UBound(m_aKeys, 1)
            nIndex = m_aKeys(i).nIndex
            
            If (nIndex >= nTargetPosition) And (nIndex < nMovedItem) Then
            m_aKeys(i).nIndex = nIndex + 1
            
            nChangedCount = nChangedCount + 1
            
            If nChangedCount = nChangeNum Then Exit For
            End If
          Next
        
        Case Else
          ' raise error here
          Err.Raise Undefined, , "Undifined error"
      End Select
      
      
      ' handle setting the index for the item that we just moved
      If m_nHoldKeyArrayIndex <> KEY_NOT_FOUND Then
        If m_nHoldKeyArrayIndex > (-1) Then m_aKeys(m_nHoldKeyArrayIndex).nIndex = nTargetPosition
      End If
      
      
      MoveItem = nTargetPosition
    End If
    
    ' reset the array index holder
    m_nHoldKeyArrayIndex = KEY_NOT_FOUND
  
  Else
    Err.Raise MoveNotAllowedWhileInEnum, , "Move not allowed while in For...Each loop."
  End If ' m_nCurrentEnumOperations = 0
  
End Function

Private Function MoveItemToEndOfArray(ByVal nItem As Long) As Long
  ' moves item to the end of the array.  this function is
  ' used to move items to be deleted from the array
  
  Dim nUbound As Long, nHoldForMove As Long
  Dim i As Long, nChangeNum As Long, nChangedCount As Long, nIndex As Long

  nUbound = UBound(m_anIndexLookp, 1)

  If nItem < nUbound Then
    ' copy the item to be moved
    nHoldForMove = m_anIndexLookp(nItem)

    ' grab all of the items above the moved one and shift them up
    CopyMemory ByVal VarPtr(m_anIndexLookp(nItem)), ByVal VarPtr(m_anIndexLookp(nItem + 1)), (nUbound - nItem) * 4&

    ' blast the moved item into the last position
    m_anIndexLookp(nUbound) = nHoldForMove
  End If

        
  ' now that we have moved the items, we need to fix the indexes held in the
  ' keys array.  we try to limit our time looping by keeping track of the
  ' number of indexes we *fix*.  once we have fixed all the indexes that need to
  ' be fixed, we don't need to loop any more.
  nChangeNum = nUbound - nItem
  
  For i = 0 To UBound(m_aKeys, 1)
    nIndex = m_aKeys(i).nIndex
    
    If (nIndex > nItem) And (nIndex <= nUbound) Then
      m_aKeys(i).nIndex = nIndex - 1
      
      nChangedCount = nChangedCount + 1
      
      If nChangedCount = nChangeNum Then Exit For
    End If
  Next


  MoveItemToEndOfArray = nUbound
  
End Function

' this should actually go in the class that wraps the collection object
' but since we didn't wrap the collection for this demo, we stuck it here.
Friend Sub FireEvent(ByVal sKey$, ByVal lpObject As Long)

  Dim i As Long, nIndex As Long
  
  nIndex = FindKey(sKey)
  
  If nIndex = KEY_NOT_FOUND Then
    ' no key was entered for this item when it was added.  we have to loop through the array and match the pointer...
    For i = m_nBaseIndex To UBound(m_anIndexLookp, 1)
      If ObjPtr(m_aItems(m_anIndexLookp(i)).oItem) = lpObject Then
        nIndex = i
        Exit For
      End If
    Next
  End If
  
'  If nIndex <> KEY_NOT_FOUND Then RaiseEvent GenericEvent(nIndex, sKey)
  
End Sub

Private Sub IDirtyKeySupport_Clean()
    Call Me.Clean
End Sub

Private Property Get IDirtyKeySupport_DirtyKeys() As KeyCollection
    Set IDirtyKeySupport_DirtyKeys = Me.DirtyKeys
End Property

Private Property Get IDirtyKeySupport_DirtyValues() As ValueCollection
    Set IDirtyKeySupport_DirtyValues = Me.DirtyValues
End Property

Private Function IDirtyKeySupport_GetEnumerator() As IDictionaryEnumerator
    Set IDirtyKeySupport_GetEnumerator = Me.GetDictionaryEnumerator
End Function

Private Function IDirtyKeySupport_IsDirty(Optional ByVal Key As Variant) As Boolean
    IDirtyKeySupport_IsDirty = Me.IsDirty(Key)
End Function

Private Function IEnumerable_GetEnumerator() As IEnumerator
    Set IEnumerable_GetEnumerator = Me.GetEnumerator
End Function
