VERSION 1.0 CLASS
BEGIN
  MultiUse = -1  'True
  Persistable = 0  'NotPersistable
  DataBindingBehavior = 0  'vbNone
  DataSourceBehavior  = 0  'vbNone
  MTSTransactionMode  = 0  'NotAnMTSObject
END
Attribute VB_Name = "Math"
Attribute VB_GlobalNameSpace = True
Attribute VB_Creatable = True
Attribute VB_PredeclaredId = False
Attribute VB_Exposed = True
Option Explicit

Private Type TYPE_DECIMAL
    Type                 As Integer
    Scale                     As Byte             ' // The number of decimal places for the
                                                  ' // number. Valid values are from 0 to 28. So
                                                  ' // 12.345 is represented as 12345 with a
                                                  ' // scale of 3.
    Sign                      As Byte             ' // 0 for positive numbers or DECIMAL_NEG for
                                                  ' // negative numbers. So -1 is represented as
                                                  ' // 1 with the DECIMAL_NEG bit set.
    Hi32                      As Long ' unsigned
    Lo32                      As Long ' unsigned
    Mid32                     As Long ' unsigned
End Type

Private varPiExcact As Variant

Public Function Acos(ByVal D As Double) As Double
    Acos = (PiExcact / 2) - Atn(D / (Sqr(1 - D ^ 2)))
End Function

Public Function Asin(ByVal D As Double) As Double
    Asin = VBA.Math.Atn(D / (Sqr(1 - D ^ 2)))
End Function

Public Function Acot(ByVal X As Double) As Double
    Acot = (PiExcact / 2) - VBA.Math.Atn(X)
End Function

Public Function Atan2(ByVal Y As Double, ByVal X As Double) As Double
    Atan2 = VBA.Math.Atn(Y / X)
End Function

Public Function Cosh(ByVal Value As Double) As Double
    Cosh = (VBA.Math.Exp(Value) + VBA.Math.Exp(-Value)) / 2
End Function

Public Function GetHighestBit(ByVal Number As Long) As Long
    If Number <> 0 Then
        GetHighestBit = 2 ^ CLng(Log(Number) / Log(2))
    End If
End Function

Public Function GetBitArray(ByVal Number As Variant) As Byte()
    Dim Bytes() As Byte
    Dim i As Long
    Dim j As Long
    
    Number = CDec(Number)
    
    ReDim Bytes(95)
    
    For i = 1 To 96
        If mdlFunctions.AndEx(Number, mdlFunctions.PowOf2(i - 1)) > 0 Then
            Bytes(j) = i
            
            j = j + 1
        End If
    Next i
    
    If j = 0 Then
        Call InitArray(Bytes, True)
    ElseIf j < 96 Then
        ReDim Preserve Bytes(j - 1)
    End If
    
    GetBitArray = Bytes()
End Function

Public Function Break(ByRef Number As Variant) As Variant
    Break = Number - Int(Number)
End Function

Public Function LoByte(ByVal wParam As Integer) As Byte
    LoByte = wParam And &HFF
End Function

Public Function HiByte(ByVal wParam As Integer) As Byte
    HiByte = (wParam And &HFF00&) \ 256
End Function

Public Function HiWord(ByVal lParam As Long) As Long
'===========================================================================
'   HiWord - Returns Hi Word part of the supplied dual long value which is often needed
'   when working with long values returned by API calls.
'
'   lParam      Input. Dual value returned by API call.
'
'   RETURNS     the high order word of the supplied long value
'===========================================================================
    HiWord = (lParam And &HFFFF0000) \ &H10000

End Function
   

Public Function LoWord(ByVal lParam As Long) As Long
'===========================================================================
'   LoWord - Returns Lo Word part of the supplied dual long value which is often needed
'   when working with long values returned by API calls.
'
'   lParam      Input. Dual value returned by API call.
'
'   RETURNS     the low order word of the supplied long value
'===========================================================================
    If (lParam And &H8000&) <> 0 Then
       LoWord = lParam Or &HFFFF0000
    Else
       LoWord = lParam And &HFFFF&
    End If
      
End Function

Function MakeWord(ByVal LoByte As Byte, ByVal HiByte As Byte) As Integer
'===========================================================================
'   MakeWord - Packs 2 8-bit integers into a 16-bit integer.
'===========================================================================

    If (HiByte And &H80) <> 0 Then
        MakeWord = ((HiByte * 256&) + LoByte) Or &HFFFF0000
    Else
        MakeWord = (HiByte * 256) + LoByte
    End If
    
End Function

Public Function MakeDWord(ByVal LoWord As Integer, ByVal HiWord As Integer) As Long
'===========================================================================
'   MakeDWord - Packs 2 16-bit integers into a 32-bit Long.
'
'   HiWord          The most significant 16 bits.
'
'   LoWord          The least significant 16 bits.
'
'   RETURNS:        A Long holding the full 32-bit value.
'===========================================================================
    
    MakeDWord = (HiWord * 65536) + (LoWord And &HFFFF&)

End Function

Public Function MakeDecimal( _
  ByVal HighDword As Long, _
  ByVal LowDword As Long, _
  Optional ByVal HighestDWord As Long, _
  Optional ByVal DecimalPlaces As Byte, _
  Optional ByVal Signed As Boolean) As Variant
'===========================================================================
' MakeDecimal - Convert up to three 32 bit integers into a 64 bit value.
'===========================================================================

  Const DECIMAL_NEG = &H80&
  Dim dec As TYPE_DECIMAL

  ' NOTE: The Decimal data type uses all of the VARIANT
  ' structure except the 2 lead VARTYPE bytes. The Decimal
  ' data type is a 12 byte/96-bit value.

  ' Fill the decimal structure
  With dec
    .Type = VT_DECIMAL
    .Lo32 = LowDword
    .Mid32 = HighDword
    .Hi32 = HighestDWord
    Select Case DecimalPlaces
    Case 0 To 28
      .Scale = DecimalPlaces
    Case Else
      Err.Raise 5
    End Select
    
    If Signed Then
      .Sign = DECIMAL_NEG
    End If
  End With
  
  ' Copy the data into return value
  RtlMoveMemory ByVal VarPtr(MakeDecimal), _
    ByVal VarPtr(dec.Type), LenB(dec)

End Function

Public Function HiDWord(ByVal vDecimal As Variant) As Long
'===========================================================================
'   HiDWord - Returns the middle 32 bits of a 64 bit value [expressed as a
'   Decimal or Currency within a Variant].
'
'   NOTE: This function is often called when working with file system based API
'   calls. For that reason we force the "value" as a Decimal/Variant. This is to
'   avoid the hassle and confusion of scaling when using a Currency data type.
'   The technique of using the Currency data type in place of structures like
'   FILETIME and LARGE_INTEGER, et al., can be very tricky and is sometimes hard
'   to tell when and why they need to be scaled and de-scaled.
'
'   vDecimal       The 64 bit value.
'
'   RETURNS     The high order DWORD of the supplied 64 bit value.
'===========================================================================

  Dim dec As TYPE_DECIMAL
  
  ' Ensure it's of the Decimal data type
  Select Case VarType(vDecimal)
  Case vbDecimal
      ' correct data type
  Case vbCurrency, vbLong, vbInteger
      ' a valid 64 bit integer
      vDecimal = CDec(vDecimal)
  Case Else
      ' anything else would be uncivilized
      Err.Raise 5
  End Select

  ' File the DECIMAL structure
  RtlMoveMemory ByVal VarPtr(dec), ByVal VarPtr(vDecimal), LenB(dec)

  HiDWord = dec.Mid32

End Function

Public Function LoDWord(ByVal vDecimal As Variant) As Long
'===========================================================================
'   LoDWord - Returns the lower 32 bits of a 64 bit value [expressed as a
'   Decimal or Currency within a Variant].
'
'   NOTE: This function is often called when working with file system based API
'   calls. For that reason we force the "value" as a Decimal/Variant. This is to
'   avoid the hassle and confusion of scaling when using a Currency data type.
'   The technique of using the Currency data type in place of structures like
'   FILETIME and LARGE_INTEGER, et al., can be very tricky and is sometimes hard
'   to tell when and why they need to be scaled and de-scaled.
'
'   vDecimal       The 64 bit value.
'
'   RETURNS     The low order DWORD of the supplied 64 bit value.
'===========================================================================
    
  Dim dec As TYPE_DECIMAL
  
  ' Ensure it's of the Decimal data type
  Select Case VarType(vDecimal)
  Case vbDecimal
      ' correct data type
  Case vbCurrency, vbLong, vbInteger
      ' a valid 64 bit integer
      vDecimal = CDec(vDecimal)
  Case Else
      ' anything else would be uncivilized
      Err.Raise 5
  End Select

  ' Fill the DECIMAL structure
  RtlMoveMemory ByVal VarPtr(dec), ByVal VarPtr(vDecimal), LenB(dec)

  LoDWord = dec.Lo32

End Function

Public Function HighestDWord(ByVal vDecimal As Variant) As Long
'===========================================================================
'   HighestDWord - Returns the Highest 32 bits of a 64 bit value [expressed
'   as a Decimal or Currency within a Variant].
'
'   NOTE: This function is often called when working with file system based API
'   calls. For that reason we force the "value" as a Decimal/Variant. This is to
'   avoid the hassle and confusion of scaling when using a Currency data type.
'   The technique of using the Currency data type in place of structures like
'   FILETIME and LARGE_INTEGER, et al., can be very tricky and is sometimes hard
'   to tell when and why they need to be scaled and de-scaled.
'
'   vDecimal      The 64 bit value.
'
'   RETURNS       The low order DWORD of the supplied 64 bit value.
'===========================================================================
    
  Dim dec As TYPE_DECIMAL
  
  ' Ensure it's of the Decimal data type
  Select Case VarType(vDecimal)
  Case vbDecimal
      ' correct data type
  Case vbCurrency, vbLong, vbInteger
      ' a valid 64 bit integer
      vDecimal = CDec(vDecimal)
  Case Else
      ' anything else would be uncivilized
      Err.Raise 5
  End Select

  ' Fill the DECIMAL structure
  RtlMoveMemory ByVal VarPtr(dec), ByVal VarPtr(vDecimal), LenB(dec)

  HighestDWord = dec.Hi32

End Function

Public Sub SplitDecimal( _
        ByVal vDecimal As Variant, _
        ByRef LowDword As Long, _
        ByRef HighDword As Long, _
        Optional ByRef HighestDWord As Long)

  Dim dec As TYPE_DECIMAL
  
  ' Ensure it's of the Decimal data type
  Select Case VarType(vDecimal)
  Case vbDecimal
      ' correct data type
  Case vbCurrency, vbLong, vbInteger
      ' a valid 64 bit integer
      vDecimal = CDec(vDecimal)
  Case Else
      ' anything else would be uncivilized
      Err.Raise 5
  End Select

  ' Fill the DECIMAL structure
  RtlMoveMemory ByVal VarPtr(dec), ByVal VarPtr(vDecimal), LenB(dec)

  ' Return each part of the decimal value
  LowDword = dec.Lo32
  HighDword = dec.Mid32
  HighestDWord = dec.Hi32

End Sub

Public Property Get Eul() As Single
    Eul = 2.71828
End Property

Public Property Get Pi() As Single
    Pi = 3.14159
End Property

Public Property Get PiExcact() As Variant
    If IsEmpty(varPiExcact) Then
        varPiExcact = CDec("3.1415926535897932384626433832795")
    End If
    
    PiExcact = varPiExcact
End Property

Public Function DivRem(ByVal Value1 As Variant, ByVal Value2 As Variant, ByRef Remainder As Long) As Variant
    DivRem = Value1 \ Value2
    Remainder = Value1 Mod Value2
End Function

Public Function Min(ByRef Value1 As Variant, ByRef Value2 As Variant) As Variant
    If Value1 < Value2 Then
        Min = Value1
    Else
        Min = Value2
    End If
End Function

Public Function Max(ByRef Value1 As Variant, ByRef Value2 As Variant) As Variant
    If Value1 > Value2 Then
        Max = Value1
    Else
        Max = Value2
    End If
End Function

Public Function Base2Log(ByVal Number As Double) As Double
    Base2Log = Log(Number) / Log(2)
End Function

Public Function Log10(ByVal Number As Double) As Double
    Log10 = Log(Number) / Log(10)
End Function

Public Function Floor(ByRef Number As Variant) As Variant
    Floor = Int(Number)
End Function

Public Function Ceiling(ByRef Number As Variant) As Variant
    Ceiling = Int(Number) + 1
End Function

Public Function XorEx(ByVal Var1 As Variant, ByVal Var2 As Variant) As Variant
    XorEx = mdlFunctions.XorEx(Var1, Var2)
End Function

Public Function OrEx(ByVal Var1 As Variant, ByVal Var2 As Variant) As Variant
    OrEx = mdlFunctions.OrEx(Var1, Var2)
End Function

Public Function AndEx(ByVal Var1 As Variant, ByVal Var2 As Variant) As Variant
    AndEx = mdlFunctions.AndEx(Var1, Var2)
End Function

Public Function NotEx(ByVal var As Variant) As Variant
    NotEx = mdlFunctions.NotEx(var)
End Function

Public Function GetBits( _
    ByVal Bits As Variant, _
    Optional ByVal StartBit As Byte = 1, _
    Optional ByVal BitCount As Byte _
    ) As Variant

    GetBits = mdlFunctions.GetBits(Bits, StartBit, BitCount)
End Function

Public Function GetBitsByOrdinal( _
    ByVal Bits As Variant, _
    Optional ByVal StartBit As Byte = 0, _
    Optional ByVal EndBit As Byte = 0 _
    ) As Variant

    GetBitsByOrdinal = mdlFunctions.GetBitsByOrdinal(Bits, StartBit, EndBit)
End Function

'Public Function GetBitsByOrdinalReverse( _
'    ByVal Bits As Variant, _
'    Optional ByVal StartBit As Byte = 1, _
'    Optional ByVal EndBit As Byte _
'    ) As Variant
'
'    GetBits = mdlFunctions.GetBits(Bits, StartBit, EndBit - StartBit + 1)
'End Function

Public Function GetBitMask( _
        Optional ByVal StartBit As Byte = 1, _
        Optional ByVal BitCount As Byte _
        ) As Variant
    GetBitMask = mdlFunctions.GetBitMask(StartBit, BitCount)
End Function

Public Function PowOf2( _
        ByVal Power As Byte _
        ) As Variant
    PowOf2 = mdlFunctions.PowOf2(Power)
End Function

Public Function Pow(ByVal Value As Double, ByVal Power As Byte) As Double
    Dim i As Long
    
    Pow = Value
    
    For i = 0 To Power
        Pow = Pow * Value
    Next
End Function

Public Static Function BankersRound( _
    ByVal Value As Variant, _
    Optional ByVal Digits As Integer _
  ) As Variant
  
  ' By Filipe Lage
  ' fclage@garciaresende.com
  ' msn: fclage@clix.pt
  ' Revision C by Donald - 20060201
  Dim xint As Double, yint As Double, xrest As Double
  Static PreviousValue    As Double
  Static PreviousDecimals As Long
  Static PreviousOutput   As Double
  Static m                As Double
  
  If m = 0 Then m = 1
    ' Initialization - M is never 0 (it is always 10 ^ n)
  
  If PreviousValue = Value And PreviousDecimals = Digits Then BankersRound = PreviousOutput: Exit Function
    ' Hey... it's the same number and decimals as before...
    ' So, the actual result is the same. No need to recalc it
  
  If Value = 0 Then Exit Function
    ' no matter what rounding is made, 0 is always rounded to 0
      
  If PreviousDecimals = Digits Then
    Else
    ' A different number of decimal places, means a new Multiplier
    PreviousDecimals = Digits
    m = 10 ^ Digits
    End If
  
  If m = 1 Then xint = Value Else xint = Value * CDec(m)
      ' Let's consider the multiplication of the number by the multiplier
      ' Bug fixed: If you just multiplied the value by M, those nasty reals came up
      ' So, we use CDEC(m) to avoid that
                                                              
  BankersRound = Fix(xint)
      ' The real integer of the number (unlike INT, FIX reports the actual number)
  
  ' 20060201: fix by Donald
  If Abs(Fix(10 * (xint - BankersRound))) > 4 Then
    If BankersRound >= 0 Then
      BankersRound = BankersRound + 1
    Else
      BankersRound = BankersRound - 1
    End If
  End If
      ' First decimal is 5 or bigger ? If so, we'll add +1 or -1 to the result (later to be divided by M)
  
  If m = 1 Then Else BankersRound = BankersRound / m
      ' Divides by the multiplier. But we only need to devide if M isn't 1
  
  PreviousOutput = BankersRound
  PreviousValue = Value
      ' Let's save this last result in memory... may be handy ;)

'  Dim i As Long
'  Dim Pot10(-28 To 28) As Variant
'
'  'Ggf. 10er-Potenzen vor-berechnen:
'  If i = 0 Then
'    For i = LBound(Pot10) To UBound(Pot10)
'      Pot10(i) = CDec(10 ^ i)
'    Next i
'  End If
'
'  'Los gehts:
'  On Error Resume Next
'
'    If Value > 0 Then
'      BankersRound = Int(Value * Pot10(Digits) + 0.5) * Pot10(-Digits)
'    Else
'      BankersRound = -Int(-Value * Pot10(Digits) + 0.5) * Pot10(-Digits)
'    End If
'    If Err.Number Then BankersRound = Value
End Function

'Public Function Round17(ByRef v As Double, Optional ByVal lngDecimals As Long = 0) As Double
'  ' By Filipe Lage
'  ' fclage@garciaresende.com
'  ' msn: fclage@clix.pt
'  ' Revision C by Donald - 20060201
'  Dim xint As Double, yint As Double, xrest As Double
'  Static PreviousValue    As Double
'  Static PreviousDecimals As Long
'  Static PreviousOutput   As Double
'  Static m                As Double
'
'  If m = 0 Then m = 1
'    ' Initialization - M is never 0 (it is always 10 ^ n)
'
'  If PreviousValue = v And PreviousDecimals = lngDecimals Then Round17 = PreviousOutput: Exit Function
'    ' Hey... it's the same number and decimals as before...
'    ' So, the actual result is the same. No need to recalc it
'
'  If v = 0 Then Exit Function
'    ' no matter what rounding is made, 0 is always rounded to 0
'
'  If PreviousDecimals = lngDecimals Then
'    Else
'    ' A different number of decimal places, means a new Multiplier
'    PreviousDecimals = lngDecimals
'    m = 10 ^ lngDecimals
'    End If
'
'  If m = 1 Then xint = v Else xint = v * CDec(m)
'      ' Let's consider the multiplication of the number by the multiplier
'      ' Bug fixed: If you just multiplied the value by M, those nasty reals came up
'      ' So, we use CDEC(m) to avoid that
'
'  Round17 = Fix(xint)
'      ' The real integer of the number (unlike INT, FIX reports the actual number)
'
'  ' 20060201: fix by Donald
'  If Abs(Fix(10 * (xint - Round17))) > 4 Then
'    If Round17 >= 0 Then
'      Round17 = Round17 + 1
'    Else
'      Round17 = Round17 - 1
'    End If
'  End If
'      ' First decimal is 5 or bigger ? If so, we'll add +1 or -1 to the result (later to be divided by M)
'
'  If m = 1 Then Else Round17 = Round17 / m
'      ' Divides by the multiplier. But we only need to devide if M isn't 1
'
'  PreviousOutput = Round17
'  PreviousValue = v
'      ' Let's save this last result in memory... may be handy ;)
'End Function

Public Function GetPercentOfBaseValue( _
        ByVal Value As Double, _
        ByVal Base As Double) _
        As Double
    GetPercentOfBaseValue = Value / Base * 100
End Function

Public Function GetPercentValue( _
        ByVal Base As Double, _
        ByVal Percent As Double) _
        As Double
    GetPercentValue = Base * Percent / 100
End Function

Public Function AddPercent( _
        ByVal Base As Double, _
        ByVal Percent As Double) _
        As Double
    AddPercent = Base * (1 + Percent / 100)
End Function

Public Function ShiftRight( _
    ByVal Value As Long, _
    ByVal ShiftCount As Long _
  ) As Long

  Select Case ShiftCount
  Case 0&:  ShiftRight = Value
  Case 1&:  ShiftRight = (Value And &HFFFFFFFE) \ &H2&
  Case 2&:  ShiftRight = (Value And &HFFFFFFFC) \ &H4&
  Case 3&:  ShiftRight = (Value And &HFFFFFFF8) \ &H8&
  Case 4&:  ShiftRight = (Value And &HFFFFFFF0) \ &H10&
  Case 5&:  ShiftRight = (Value And &HFFFFFFE0) \ &H20&
  Case 6&:  ShiftRight = (Value And &HFFFFFFC0) \ &H40&
  Case 7&:  ShiftRight = (Value And &HFFFFFF80) \ &H80&
  Case 8&:  ShiftRight = (Value And &HFFFFFF00) \ &H100&
  Case 9&:  ShiftRight = (Value And &HFFFFFE00) \ &H200&
  Case 10&: ShiftRight = (Value And &HFFFFFC00) \ &H400&
  Case 11&: ShiftRight = (Value And &HFFFFF800) \ &H800&
  Case 12&: ShiftRight = (Value And &HFFFFF000) \ &H1000&
  Case 13&: ShiftRight = (Value And &HFFFFE000) \ &H2000&
  Case 14&: ShiftRight = (Value And &HFFFFC000) \ &H4000&
  Case 15&: ShiftRight = (Value And &HFFFF8000) \ &H8000&
  Case 16&: ShiftRight = (Value And &HFFFF0000) \ &H10000
  Case 17&: ShiftRight = (Value And &HFFFE0000) \ &H20000
  Case 18&: ShiftRight = (Value And &HFFFC0000) \ &H40000
  Case 19&: ShiftRight = (Value And &HFFF80000) \ &H80000
  Case 20&: ShiftRight = (Value And &HFFF00000) \ &H100000
  Case 21&: ShiftRight = (Value And &HFFE00000) \ &H200000
  Case 22&: ShiftRight = (Value And &HFFC00000) \ &H400000
  Case 23&: ShiftRight = (Value And &HFF800000) \ &H800000
  Case 24&: ShiftRight = (Value And &HFF000000) \ &H1000000
  Case 25&: ShiftRight = (Value And &HFE000000) \ &H2000000
  Case 26&: ShiftRight = (Value And &HFC000000) \ &H4000000
  Case 27&: ShiftRight = (Value And &HF8000000) \ &H8000000
  Case 28&: ShiftRight = (Value And &HF0000000) \ &H10000000
  Case 29&: ShiftRight = (Value And &HE0000000) \ &H20000000
  Case 30&: ShiftRight = (Value And &HC0000000) \ &H40000000
  Case 31&: ShiftRight = CBool(Value And &H80000000)
  End Select
End Function

Public Function ShiftLeft( _
    ByVal Value As Long, _
    ByVal ShiftCount As Long _
  ) As Long
  
  Select Case ShiftCount
  Case 0&
    ShiftLeft = Value
  Case 1&
    If Value And &H40000000 Then
      ShiftLeft = (Value And &H3FFFFFFF) * &H2& Or &H80000000
    Else
      ShiftLeft = (Value And &H3FFFFFFF) * &H2&
    End If
  Case 2&
    If Value And &H20000000 Then
      ShiftLeft = (Value And &H1FFFFFFF) * &H4& Or &H80000000
    Else
      ShiftLeft = (Value And &H1FFFFFFF) * &H4&
    End If
  Case 3&
    If Value And &H10000000 Then
      ShiftLeft = (Value And &HFFFFFFF) * &H8& Or &H80000000
    Else
      ShiftLeft = (Value And &HFFFFFFF) * &H8&
    End If
  Case 4&
    If Value And &H8000000 Then
      ShiftLeft = (Value And &H7FFFFFF) * &H10& Or &H80000000
    Else
      ShiftLeft = (Value And &H7FFFFFF) * &H10&
    End If
  Case 5&
    If Value And &H4000000 Then
      ShiftLeft = (Value And &H3FFFFFF) * &H20& Or &H80000000
    Else
      ShiftLeft = (Value And &H3FFFFFF) * &H20&
    End If
  Case 6&
    If Value And &H2000000 Then
      ShiftLeft = (Value And &H1FFFFFF) * &H40& Or &H80000000
    Else
      ShiftLeft = (Value And &H1FFFFFF) * &H40&
    End If
  Case 7&
    If Value And &H1000000 Then
      ShiftLeft = (Value And &HFFFFFF) * &H80& Or &H80000000
    Else
      ShiftLeft = (Value And &HFFFFFF) * &H80&
    End If
  Case 8&
    If Value And &H800000 Then
      ShiftLeft = (Value And &H7FFFFF) * &H100& Or &H80000000
    Else
      ShiftLeft = (Value And &H7FFFFF) * &H100&
    End If
  Case 9&
    If Value And &H400000 Then
      ShiftLeft = (Value And &H3FFFFF) * &H200& Or &H80000000
    Else
      ShiftLeft = (Value And &H3FFFFF) * &H200&
    End If
  Case 10&
    If Value And &H200000 Then
      ShiftLeft = (Value And &H1FFFFF) * &H400& Or &H80000000
    Else
      ShiftLeft = (Value And &H1FFFFF) * &H400&
    End If
  Case 11&
    If Value And &H100000 Then
      ShiftLeft = (Value And &HFFFFF) * &H800& Or &H80000000
    Else
      ShiftLeft = (Value And &HFFFFF) * &H800&
    End If
  Case 12&
    If Value And &H80000 Then
      ShiftLeft = (Value And &H7FFFF) * &H1000& Or &H80000000
    Else
      ShiftLeft = (Value And &H7FFFF) * &H1000&
    End If
  Case 13&
    If Value And &H40000 Then
      ShiftLeft = (Value And &H3FFFF) * &H2000& Or &H80000000
    Else
      ShiftLeft = (Value And &H3FFFF) * &H2000&
    End If
  Case 14&
    If Value And &H20000 Then
      ShiftLeft = (Value And &H1FFFF) * &H4000& Or &H80000000
    Else
      ShiftLeft = (Value And &H1FFFF) * &H4000&
    End If
  Case 15&
    If Value And &H10000 Then
      ShiftLeft = (Value And &HFFFF&) * &H8000& Or &H80000000
    Else
      ShiftLeft = (Value And &HFFFF&) * &H8000&
    End If
  Case 16&
    If Value And &H8000& Then
      ShiftLeft = (Value And &H7FFF&) * &H10000 Or &H80000000
    Else
      ShiftLeft = (Value And &H7FFF&) * &H10000
    End If
  Case 17&
    If Value And &H4000& Then
      ShiftLeft = (Value And &H3FFF&) * &H20000 Or &H80000000
    Else
      ShiftLeft = (Value And &H3FFF&) * &H20000
    End If
  Case 18&
    If Value And &H2000& Then
      ShiftLeft = (Value And &H1FFF&) * &H40000 Or &H80000000
    Else
      ShiftLeft = (Value And &H1FFF&) * &H40000
    End If
  Case 19&
    If Value And &H1000& Then
      ShiftLeft = (Value And &HFFF&) * &H80000 Or &H80000000
    Else
      ShiftLeft = (Value And &HFFF&) * &H80000
    End If
  Case 20&
    If Value And &H800& Then
      ShiftLeft = (Value And &H7FF&) * &H100000 Or &H80000000
    Else
      ShiftLeft = (Value And &H7FF&) * &H100000
    End If
  Case 21&
    If Value And &H400& Then
      ShiftLeft = (Value And &H3FF&) * &H200000 Or &H80000000
    Else
      ShiftLeft = (Value And &H3FF&) * &H200000
    End If
  Case 22&
    If Value And &H200& Then
      ShiftLeft = (Value And &H1FF&) * &H400000 Or &H80000000
    Else
      ShiftLeft = (Value And &H1FF&) * &H400000
    End If
  Case 23&
    If Value And &H100& Then
      ShiftLeft = (Value And &HFF&) * &H800000 Or &H80000000
    Else
      ShiftLeft = (Value And &HFF&) * &H800000
    End If
  Case 24&
    If Value And &H80& Then
      ShiftLeft = (Value And &H7F&) * &H1000000 Or &H80000000
    Else
      ShiftLeft = (Value And &H7F&) * &H1000000
    End If
  Case 25&
    If Value And &H40& Then
      ShiftLeft = (Value And &H3F&) * &H2000000 Or &H80000000
    Else
      ShiftLeft = (Value And &H3F&) * &H2000000
    End If
  Case 26&
    If Value And &H20& Then
      ShiftLeft = (Value And &H1F&) * &H4000000 Or &H80000000
    Else
      ShiftLeft = (Value And &H1F&) * &H4000000
    End If
  Case 27&
    If Value And &H10& Then
      ShiftLeft = (Value And &HF&) * &H8000000 Or &H80000000
    Else
      ShiftLeft = (Value And &HF&) * &H8000000
    End If
  Case 28&
    If Value And &H8& Then
      ShiftLeft = (Value And &H7&) * &H10000000 Or &H80000000
    Else
      ShiftLeft = (Value And &H7&) * &H10000000
    End If
  Case 29&
    If Value And &H4& Then
      ShiftLeft = (Value And &H3&) * &H20000000 Or &H80000000
    Else
      ShiftLeft = (Value And &H3&) * &H20000000
    End If
  Case 30&
    If Value And &H2& Then
      ShiftLeft = (Value And &H1&) * &H40000000 Or &H80000000
    Else
      ShiftLeft = (Value And &H1&) * &H40000000
    End If
  Case 31&
    If Value And &H1& Then
      ShiftLeft = &H80000000
    Else
      ShiftLeft = &H0&
    End If
  End Select
End Function
